#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include "1_dwarf2.h"
#include "1_const.h"

#define START 0x400259
#define END   0x403217

#define TRACE 1
#define TRACE_CURRENT_PTR 1

unsigned char *old_op_ptr;
#define trace_print(fmt, ...) \
    do { \
        if (TRACE_CURRENT_PTR) printf("# %p\n", old_op_ptr); \
        if (TRACE) printf(fmt, ##__VA_ARGS__); \
    } while (0)

char *data = NULL;

union unaligned {
    void *p;
    unsigned u2 __attribute__ ((mode (HI)));
    unsigned u4 __attribute__ ((mode (SI)));
    unsigned u8 __attribute__ ((mode (DI)));
    signed s2 __attribute__ ((mode (HI)));
    signed s4 __attribute__ ((mode (SI)));
    signed s8 __attribute__ ((mode (DI)));
} __attribute__ ((packed));

typedef struct _instrumented_value {
    unsigned int id;
    unsigned int dirty;
    _Unwind_Word value;
} _instrumented_value;

const void * _read(const void *p) { return (const void *)(&dwarfdwarf[(long)p - START]); }

static inline void *
read_pointer (const void *p) { const union unaligned *up = _read(p) ; return up->p; }

static inline int
read_1u (const void *p) { return *(const unsigned char *) _read(p); }

static inline int
read_1s (const void *p) { return *(const signed char *) _read(p); }

static inline int
read_2u (const void *p) { const union unaligned *up = _read(p); return up->u2; }

static inline int
read_2s (const void *p) { const union unaligned *up = _read(p); return up->s2; }

static inline unsigned int
read_4u (const void *p) { const union unaligned *up = _read(p); return up->u4; }

static inline int
read_4s (const void *p) { const union unaligned *up = _read(p); return up->s4; }

static inline unsigned long
read_8u (const void *p) { const union unaligned *up = _read(p); return up->u8; }

static inline unsigned long
read_8s (const void *p) { const union unaligned *up = _read(p); return up->s8; }


/* Read an unsigned leb128 value from P, store the value in VAL, return
   P incremented past the value.  We assume that a word is large enough to
   hold any value so encoded; if it is smaller than a pointer on some target,
   pointers should not be leb128 encoded on that target.  */

static const unsigned char *
read_uleb128 (const unsigned char *p, _uleb128_t *val)
{
    unsigned int shift = 0;
    unsigned char byte;
    _uleb128_t result;

    result = 0;
    do
    {
        byte = read_1u(p++);
        result |= ((_uleb128_t)byte & 0x7f) << shift;
        shift += 7;
    }
    while (byte & 0x80);

    *val = result;
    return p;
}

static const unsigned char *
read_sleb128 (const unsigned char *p, _sleb128_t *val)
{
    unsigned int shift = 0;
    unsigned char byte;
    _uleb128_t result;

    result = 0;
    do
    {
        byte = read_1u(p++);
        result |= ((_uleb128_t)byte & 0x7f) << shift;
        shift += 7;
    }
    while (byte & 0x80);

    /* Sign-extend a negative value.  */
    if (shift < 8 * sizeof(result) && (byte & 0x40) != 0)
        result |= -(((_uleb128_t)1L) << shift);

    *val = (_sleb128_t) result;
    return p;
}

static _Unwind_Word 
execute_stack_op (const unsigned char *op_ptr, const unsigned char *op_end,
        /* struct _Unwind_Context *context ,*/ _instrumented_value initial[], unsigned initial_size)
{
    _instrumented_value stack[64];        /* ??? Assume this is enough.  */
    unsigned int id = initial_size;
    const unsigned char *old_op_ptr = op_ptr;
    int stack_elt;
    for (stack_elt = 0; stack_elt < initial_size; stack_elt++) {
        stack[stack_elt] = initial[stack_elt];
        trace_print("id_%u = 0x%lx\n", initial[stack_elt].id, initial[stack_elt].value);
    }

    while (op_ptr < op_end)
    {
        enum dwarf_location_atom op = read_1u(op_ptr++);
        _Unwind_Word result;
        _uleb128_t reg, utmp;
        _sleb128_t offset, stmp;
        _instrumented_value t;
        unsigned int dirty = 0;
        old_op_ptr = op_ptr;

        switch (op)
        {
            case DW_OP_lit0:
            case DW_OP_lit1:
            case DW_OP_lit2:
            case DW_OP_lit3:
            case DW_OP_lit4:
            case DW_OP_lit5:
            case DW_OP_lit6:
            case DW_OP_lit7:
            case DW_OP_lit8:
            case DW_OP_lit9:
            case DW_OP_lit10:
            case DW_OP_lit11:
            case DW_OP_lit12:
            case DW_OP_lit13:
            case DW_OP_lit14:
            case DW_OP_lit15:
            case DW_OP_lit16:
            case DW_OP_lit17:
            case DW_OP_lit18:
            case DW_OP_lit19:
            case DW_OP_lit20:
            case DW_OP_lit21:
            case DW_OP_lit22:
            case DW_OP_lit23:
            case DW_OP_lit24:
            case DW_OP_lit25:
            case DW_OP_lit26:
            case DW_OP_lit27:
            case DW_OP_lit28:
            case DW_OP_lit29:
            case DW_OP_lit30:
            case DW_OP_lit31:
                result = op - DW_OP_lit0;
                break;
            case DW_OP_addr:
                result = (_Unwind_Word) (_Unwind_Ptr) read_pointer (op_ptr);
                op_ptr += sizeof (void *);
                break;
            case DW_OP_GNU_encoded_addr:
                {
                    fprintf(stderr, "Not supported");
                    return -1;
                }
                break;
            case DW_OP_const1u:
                result = read_1u (op_ptr);
                op_ptr += 1;
                break;
            case DW_OP_const1s:
                result = read_1s (op_ptr);
                op_ptr += 1;
                break;
            case DW_OP_const2u:
                result = read_2u (op_ptr);
                op_ptr += 2;
                break;
            case DW_OP_const2s:
                result = read_2s (op_ptr);
                op_ptr += 2;
                break;
            case DW_OP_const4u:
                result = read_4u (op_ptr);
                op_ptr += 4;
                break;
            case DW_OP_const4s:
                result = read_4s (op_ptr);
                op_ptr += 4;
                break;
            case DW_OP_const8u:
                result = read_8u (op_ptr);
                op_ptr += 8;
                break;
            case DW_OP_const8s:
                result = read_8s (op_ptr);
                op_ptr += 8;
                break;
            case DW_OP_constu:
                op_ptr = read_uleb128 (op_ptr, &utmp);
                result = (_Unwind_Word)utmp;
                break;
            case DW_OP_consts:
                op_ptr = read_sleb128 (op_ptr, &stmp);
                result = (_Unwind_Sword)stmp;
                break;
            case DW_OP_reg0:
            case DW_OP_reg1:
            case DW_OP_reg2:
            case DW_OP_reg3:
            case DW_OP_reg4:
            case DW_OP_reg5:
            case DW_OP_reg6:
            case DW_OP_reg7:
            case DW_OP_reg8:
            case DW_OP_reg9:
            case DW_OP_reg10:
            case DW_OP_reg11:
            case DW_OP_reg12:
            case DW_OP_reg13:
            case DW_OP_reg14:
            case DW_OP_reg15:
            case DW_OP_reg16:
            case DW_OP_reg17:
            case DW_OP_reg18:
            case DW_OP_reg19:
            case DW_OP_reg20:
            case DW_OP_reg21:
            case DW_OP_reg22:
            case DW_OP_reg23:
            case DW_OP_reg24:
            case DW_OP_reg25:
            case DW_OP_reg26:
            case DW_OP_reg27:
            case DW_OP_reg28:
            case DW_OP_reg29:
            case DW_OP_reg30:
            case DW_OP_reg31:
                fprintf(stderr, "Not supported");
                return -1;
            case DW_OP_regx:
                fprintf(stderr, "Not supported");
                return -1;
            case DW_OP_breg0:
            case DW_OP_breg1:
            case DW_OP_breg2:
            case DW_OP_breg3:
            case DW_OP_breg4:
            case DW_OP_breg5:
            case DW_OP_breg6:
            case DW_OP_breg7:
            case DW_OP_breg8:
            case DW_OP_breg9:
            case DW_OP_breg10:
            case DW_OP_breg11:
            case DW_OP_breg12:
            case DW_OP_breg13:
            case DW_OP_breg14:
            case DW_OP_breg15:
            case DW_OP_breg16:
            case DW_OP_breg17:
            case DW_OP_breg18:
            case DW_OP_breg19:
            case DW_OP_breg20:
            case DW_OP_breg21:
            case DW_OP_breg22:
            case DW_OP_breg23:
            case DW_OP_breg24:
            case DW_OP_breg25:
            case DW_OP_breg26:
            case DW_OP_breg27:
            case DW_OP_breg28:
            case DW_OP_breg29:
            case DW_OP_breg30:
            case DW_OP_breg31:
                fprintf(stderr, "Not supported");
                return -1;
            case DW_OP_bregx:
                fprintf(stderr, "Not supported");
                return -1;
            case DW_OP_dup:
                t = stack[stack_elt - 1];
                result = t.value;
                if (t.dirty == 1) {
                    trace_print("id_%u = id_%u\n", id, t.id);
                    dirty = 1;
                }
                break;
            case DW_OP_drop:
                t = stack[stack_elt - 1];
                stack_elt -= 1;
                goto no_push;
            case DW_OP_pick:
                offset = read_1u(op_ptr++);
                t = stack[stack_elt - 1 - offset];
                result = t.value;
                if (t.dirty == 1) {
                    trace_print("id_%u = id_%u\n", id, t.id);
                    dirty = 1;
                }
                break;
            case DW_OP_over:
                t = stack[stack_elt - 2];
                result = t.value;
                if (t.dirty == 1) {
                    trace_print("id_%u = id_%u\n", id, t.id);
                    dirty = 1;
                }
                break;
            case DW_OP_swap:
                {
                    t = stack[stack_elt - 1];
                    stack[stack_elt - 1] = stack[stack_elt - 2];
                    stack[stack_elt - 2] = t;
                    goto no_push;
                }
            case DW_OP_rot:
                {
                    _instrumented_value t1, t2, t3;
                    t1 = stack[stack_elt - 1];
                    t2 = stack[stack_elt - 2];
                    t3 = stack[stack_elt - 3];
                    stack[stack_elt - 1] = t2;
                    stack[stack_elt - 2] = t3;
                    stack[stack_elt - 3] = t1;
                    goto no_push;
                }
            case DW_OP_deref:
            case DW_OP_deref_size:
            case DW_OP_abs:
            case DW_OP_neg:
            case DW_OP_not:
            case DW_OP_plus_uconst:
                /* Unary operations.  */
                stack_elt -= 1;
                t = stack[stack_elt];
                result = t.value;
                if (t.dirty == 1) {
                    dirty = 1;
                }
                switch (op)
                {
                    case DW_OP_deref:
                        {
                            void *ptr = (void *) (_Unwind_Ptr) result;
                            result = (_Unwind_Ptr) read_pointer (ptr);
                            if (t.dirty == 1) {
                                trace_print("# DEREF id_%u = *(id_%u)\n", id, t.id);
                                trace_print("id_%u = 0x%lx\n", id, result);
                            }
                        }
                        break;
                    case DW_OP_deref_size:
                        {
                            void *ptr = (void *) (_Unwind_Ptr) result;
                            int deref_size = read_1u(op_ptr++);
                            switch (deref_size)
                            {
                                case 1:
                                    result = read_1u (ptr);
                                    break;
                                case 2:
                                    result = read_2u (ptr);
                                    break;
                                case 4:
                                    result = read_4u (ptr);
                                    break;
                                case 8:
                                    result = read_8u (ptr);
                                    break;
                                default:
                                    fprintf(stderr, "Not supported");
                                    return -1;
                            }
                            if (t.dirty == 1) {
                                trace_print("# id_%u = *(id_%u) SIZE: %i\n", id, t.id, deref_size);
                                trace_print("id_%u = 0x%lx\n", id, result);
                            }
                        }
                        break;
                    case DW_OP_abs:
                        if ((_Unwind_Sword) result < 0) {
                            result = -result;
                            if (t.dirty == 1) {
                                trace_print("id_%u = - id_%u\n", id, t.id);
                            }
                        }
                        break;
                    case DW_OP_neg:
                        result = -result;
                        if (t.dirty == 1) {
                            trace_print("id_%u = - id_%u\n", id, t.id);
                        }
                        break;
                    case DW_OP_not:
                        result = ~result;
                        if (t.dirty == 1) {
                            trace_print("id_%u = not(id_%u)\n", id, t.id);
                        }
                        break;
                    case DW_OP_plus_uconst:
                        op_ptr = read_uleb128 (op_ptr, &utmp);
                        result += (_Unwind_Word)utmp;
                        if (t.dirty == 1) {
                            trace_print("id_%u = id_%u + 0x%x\n", id, t.id, utmp);
                        }
                        break;
                    default:
                        fprintf(stderr, "Not supported");
                        return -1;
                }
                break;
            case DW_OP_and:
            case DW_OP_div:
            case DW_OP_minus:
            case DW_OP_mod:
            case DW_OP_mul:
            case DW_OP_or:
            case DW_OP_plus:
            case DW_OP_shl:
            case DW_OP_shr:
            case DW_OP_shra:
            case DW_OP_xor:
            case DW_OP_le:
            case DW_OP_ge:
            case DW_OP_eq:
            case DW_OP_lt:
            case DW_OP_gt:
            case DW_OP_ne:
                {
                    /* Binary operations.  */
                    _Unwind_Word first, second;
                    stack_elt -= 2;
                    _instrumented_value sec = stack[stack_elt];
                    _instrumented_value fir = stack[stack_elt + 1];
                    second = sec.value;
                    first = fir.value;
                    if (sec.dirty == 1 || fir.dirty == 1) {
                        dirty = 1;
                    }
                    if (dirty && !sec.dirty) {
                        // has never been printed 
                        trace_print("id_%u = 0x%lx\n", sec.id, second);
                    }
                    if (dirty && !fir.dirty) {
                        // has never been printed 
                        trace_print("id_%u = 0x%lx\n", fir.id, first);
                    }

                    switch (op)
                    {
                        case DW_OP_and:
                            result = second & first;
                            if (dirty)
                                trace_print("id_%u = id_%u & id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_div:
                            result = (_Unwind_Sword) second / (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u / id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_minus:
                            result = second - first;
                            if (dirty)
                                trace_print("id_%u = id_%u - id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_mod:
                            result = second % first;
                            if (dirty)
                                trace_print("id_%u = id_%u %% id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_mul:
                            result = second * first;
                            if (dirty)
                                trace_print("id_%u = id_%u * id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_or:
                            result = second | first;
                            if (dirty)
                                trace_print("id_%u = id_%u | id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_plus:
                            result = second + first;
                            if (dirty)
                                trace_print("id_%u = id_%u + id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_shl:
                            result = second << first;
                            if (dirty)
                                trace_print("id_%u = id_%u << id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_shr:
                            result = second >> first;
                            if (dirty)
                                trace_print("id_%u = id_%u >> id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_shra:
                            result = (_Unwind_Sword) second >> first;
                            if (dirty)
                                trace_print("id_%u = (signed) id_%u >> id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_xor:
                            result = second ^ first;
                            if (dirty)
                                trace_print("id_%u = id_%u ^ id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_le:
                            result = (_Unwind_Sword) second <= (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u <= id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_ge:
                            result = (_Unwind_Sword) second >= (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u >= id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_eq:
                            result = (_Unwind_Sword) second == (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u == id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_lt:
                            result = (_Unwind_Sword) second < (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u < id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_gt:
                            result = (_Unwind_Sword) second > (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u > id_%u\n", id, sec.id, fir.id);
                            break;
                        case DW_OP_ne:
                            result = (_Unwind_Sword) second != (_Unwind_Sword) first;
                            if (dirty)
                                trace_print("id_%u = id_%u != id_%u\n", id, sec.id, fir.id);
                            break;
                        default:
                            fprintf(stderr, "Not supported");
                            return -1;
                    }
                }
                break;
            case DW_OP_skip:
                offset = read_2s (op_ptr);
                op_ptr += 2;
                op_ptr += offset;
                trace_print("# JMP to %p\n", op_ptr);
                goto no_push;
            case DW_OP_bra:
                stack_elt -= 1;
                offset = read_2s (op_ptr);
                op_ptr += 2;
                _instrumented_value t = stack[stack_elt];
                if (t.value != 0) 
                    op_ptr += offset;

                if (t.dirty == 1) { 
                    trace_print("# BRA to %p (id_%u == 0x%lx)\n", op_ptr ,t.id, t.value);
                }
                goto no_push;
            case DW_OP_nop:
                goto no_push;
            default:
                fprintf(stderr, "Not supported");
                return -1;
        }
        /* Most things push a result value.  */
        //gcc_assert ((size_t) stack_elt < sizeof(stack)/sizeof(*stack));
        _instrumented_value res = { .id = id++, .dirty = dirty, .value = result };
        stack[stack_elt++] = res;
no_push:;
    }
    /* We were executing this program to get a value.  It should be
       at top of stack.  */
    stack_elt -= 1;

    if (!stack[stack_elt].dirty)
        trace_print("id_%u = 0x%lx\n", stack[stack_elt].id, stack[stack_elt].value);
    return stack[stack_elt].value;
}

int main(int argc, char **argv) 
{
    if (argc != 2 || strlen(argv[1]) != 32) {
        printf("Usage : %s <flag>\n", argv[0]);
        return 1;
    }

    char *arg1 = &argv[1][0];
    char *arg2 = &argv[1][8];
    char *arg3 = &argv[1][16];
    char *arg4 = &argv[1][24];

    _instrumented_value val0 = { .id = 0, .dirty = 0, .value = 0x0000fffffffffbb0 };
    _instrumented_value val1 = { .id = 1, .dirty = 1, .value = *(long*) arg1 };
    _instrumented_value val2 = { .id = 2, .dirty = 1, .value = *(long*) arg2 };
    _instrumented_value val3 = { .id = 3, .dirty = 1, .value = *(long*) arg3 };
    _instrumented_value val4 = { .id = 4, .dirty = 1, .value = *(long*) arg4 };

    _instrumented_value initial_stack[] = { val0, val1, val2, val3, val4 };
    _Unwind_Word top_stack = execute_stack_op((const void *)START + 37, (const void*) END, initial_stack, sizeof(initial_stack)/sizeof(initial_stack[0]));

    if (!TRACE) 
        printf("0x%lx\n", top_stack);
    return 0;
}

